from zict.async_buffer import AsyncBuffer as AsyncBuffer
from zict.buffer import Buffer as Buffer
from zict.cache import Cache as Cache
from zict.cache import WeakValueMapping as WeakValueMapping
from zict.file import File as File
from zict.func import Func as Func
from zict.lmdb import LMDB as LMDB
from zict.lru import LRU as LRU
from zict.sieve import Sieve as Sieve
from zict.utils import InsertionSortedSet as InsertionSortedSet
from zict.zip import Zip as Zip

# Must be kept aligned with setup.cfg
__version__ = "3.0.0"



import random
import threading
from collections import UserDict
from concurrent.futures import ThreadPoolExecutor

import pytest

from zict import Buffer
from zict.tests import utils_test


def test_simple():
    a = {}
    b = {}
    buff = Buffer(a, b, n=10, weight=lambda k, v: v)

    buff["x"] = 1
    buff["y"] = 2

    assert buff["x"] == 1
    assert buff["y"] == 2
    assert a == {"x": 1, "y": 2}
    assert buff.fast.total_weight == 3

    buff["z"] = 8
    assert a == {"y": 2, "z": 8}
    assert b == {"x": 1}

    assert buff["x"] == 1
    assert a == {"x": 1, "z": 8}
    assert b == {"y": 2}

    assert "x" in buff
    assert "y" in buff
    assert "missing" not in buff

    buff["y"] = 1
    assert a == {"x": 1, "y": 1, "z": 8}
    assert buff.fast.total_weight == 10
    assert b == {}

    del buff["z"]
    assert a == {"x": 1, "y": 1}
    assert buff.fast.total_weight == 2
    assert b == {}

    del buff["y"]
    assert a == {"x": 1}
    assert buff.fast.total_weight == 1
    assert b == {}

    assert "y" not in buff

    buff["a"] = 5
    assert set(buff) == set(buff.keys()) == {"a", "x"}

    fast_keys = set(buff.fast)
    slow_keys = set(buff.slow)
    assert not (fast_keys & slow_keys)
    assert fast_keys | slow_keys == set(buff)

    # Overweight element stays in slow mapping
    buff["b"] = 1000
    assert "b" in buff.slow
    assert set(buff.fast) == fast_keys
    assert set(buff.slow) == {"b"} | slow_keys
    assert "b" in buff
    assert buff["b"] == 1000


def test_setitem_avoid_fast_slow_duplicate():
    a = {}
    b = {}
    buff = Buffer(a, b, n=10, weight=lambda k, v: v)
    for first, second in [(1, 12), (12, 1)]:
        buff["a"] = first
        assert buff["a"] == first
        buff["a"] = second
        assert buff["a"] == second

        fast_keys = set(buff.fast)
        slow_keys = set(buff.slow)
        assert not (fast_keys & slow_keys)
        assert fast_keys | slow_keys == set(buff)

        del buff["a"]
        assert "a" not in buff
        assert "a" not in a
        assert "a" not in b


def test_mapping():
    """
    Test mapping interface for Buffer().
    """
    a = {}
    b = {}
    buff = Buffer(a, b, n=2)
    utils_test.check_mapping(buff)
    utils_test.check_closing(buff)

    buff.clear()
    assert not buff.slow
    assert not buff._cancel_restore
    assert not buff.fast
    assert not buff.fast.d
    assert not buff.fast.weights
    assert not buff.fast.total_weight
    assert not buff.fast._cancel_evict


def test_callbacks():
    f2s = []

    def f2s_cb(k, v):
        f2s.append(k)

    s2f = []

    def s2f_cb(k, v):
        s2f.append(k)

    a = {}
    b = {}
    buff = Buffer(
        a,
        b,
        n=10,
        weight=lambda k, v: v,
        fast_to_slow_callbacks=f2s_cb,
        slow_to_fast_callbacks=s2f_cb,
    )

    buff["x"] = 1
    buff["y"] = 2

    assert buff["x"] == 1
    assert buff["y"] == 2
    assert not f2s
    assert not s2f

    buff["z"] = 8

    assert f2s == ["x"]
    assert s2f == []
    buff["z"]

    assert f2s == ["x"]
    assert s2f == []

    buff["x"]
    assert f2s == ["x", "y"]
    assert s2f == ["x"]


def test_callbacks_exception_catch():
    class MyError(Exception):
        pass

    f2s = []

    def f2s_cb(k, v):
        if v > 10:
            raise MyError()
        f2s.append(k)

    s2f = []

    def s2f_cb(k, v):
        s2f.append(k)

    a = {}
    b = {}
    buff = Buffer(
        a,
        b,
        n=10,
        weight=lambda k, v: v,
        fast_to_slow_callbacks=f2s_cb,
        slow_to_fast_callbacks=s2f_cb,
    )

    buff["x"] = 1
    buff["y"] = 2

    assert buff["x"] == 1
    assert buff["y"] == 2
    assert not f2s
    assert not s2f
    assert a == {"x": 1, "y": 2}  # keys are in fast/memory
    assert not b

    # Add key < n but total weight > n this will move x out of fast
    buff["z"] = 8

    assert f2s == ["x"]
    assert s2f == []
    assert a == {"y": 2, "z": 8}
    assert b == {"x": 1}

    # Add key > n, again total weight > n this will move everything to slow except w
    # that stays in fast due to callback raising
    with pytest.raises(MyError):
        buff["w"] = 11

    assert f2s == ["x", "y", "z"]
    assert s2f == []
    assert a == {"w": 11}
    assert b == {"x": 1, "y": 2, "z": 8}


def test_n_offset():
    buff = Buffer({}, {}, n=5)
    assert buff.n == 5
    assert buff.fast.n == 5
    buff.n = 3
    assert buff.fast.n == 3
    assert buff.offset == 0
    assert buff.fast.offset == 0
    buff.offset = 2
    assert buff.offset == 2
    assert buff.fast.offset == 2


def test_set_noevict():
    a = {}
    b = {}
    f2s = []
    s2f = []
    buff = Buffer(
        a,
        b,
        n=5,
        weight=lambda k, v: v,
        fast_to_slow_callbacks=lambda k, v: f2s.append(k),
        slow_to_fast_callbacks=lambda k, v: s2f.append(k),
    )
    buff.set_noevict("x", 3)
    buff.set_noevict("y", 3)  # Would cause x to move to slow
    buff.set_noevict("z", 6)  # >n; would be immediately evicted

    assert a == {"x": 3, "y": 3, "z": 6}
    assert b == {}
    assert f2s == s2f == []

    buff.evict_until_below_target()
    assert a == {"y": 3}
    assert b == {"z": 6, "x": 3}
    assert f2s == ["z", "x"]
    assert s2f == []

    # set_noevict clears slow
    f2s.clear()
    buff.set_noevict("x", 1)
    assert a == {"y": 3, "x": 1}
    assert b == {"z": 6}
    assert f2s == s2f == []

    # Custom target; 0 != None
    buff.evict_until_below_target(0)
    assert a == {}
    assert b == {"z": 6, "x": 1, "y": 3}
    assert f2s == ["y", "x"]
    assert s2f == []


def test_evict_restore_during_iter():
    """Test that __iter__ won't be disrupted if another thread evicts or restores a key"""
    buff = Buffer({"x": 1, "y": 2}, {"z": 3}, n=5)
    assert list(buff) == ["x", "y", "z"]
    it = iter(buff)
    assert next(it) == "x"
    buff.fast.evict("x")
    assert next(it) == "y"
    assert buff["x"] == 1
    assert next(it) == "z"
    with pytest.raises(StopIteration):
        next(it)


@pytest.mark.parametrize("event", ("set", "set_noevict", "del"))
@pytest.mark.parametrize("when", ("before", "after"))
def test_cancel_evict(event, when):
    """See also:

    test_cancel_restore
    test_lru.py::test_cancel_evict
    """
    ev1 = threading.Event()
    ev2 = threading.Event()

    class Slow(UserDict):
        def __setitem__(self, k, v):
            if when == "before":
                ev1.set()
                assert ev2.wait(timeout=5)
                super().__setitem__(k, v)
            else:
                super().__setitem__(k, v)
                ev1.set()
                assert ev2.wait(timeout=5)

    buff = Buffer({}, Slow(), n=100, weight=lambda k, v: v)
    buff.set_noevict("x", 1)
    with ThreadPoolExecutor(1) as ex:
        fut = ex.submit(buff.fast.evict)
        assert ev1.wait(timeout=5)
        # cb is running

        if event == "set":
            buff["x"] = 2
        elif event == "set_noevict":
            buff.set_noevict("x", 2)
        else:
            assert event == "del"
            del buff["x"]
        ev2.set()
        assert fut.result() == (None, None, 0)

    if event in ("set", "set_noevict"):
        assert buff.fast == {"x": 2}
        assert not buff.slow
        assert buff.fast.weights == {"x": 2}
        assert list(buff.fast.order) == ["x"]
    else:
        assert not buff.fast
        assert not buff.slow
        assert not buff.fast.weights
        assert not buff.fast.order

    assert not buff.fast._cancel_evict


@pytest.mark.parametrize("event", ("set", "set_noevict", "del"))
@pytest.mark.parametrize("when", ("before", "after"))
def test_cancel_restore(event, when):
    """See also:

    test_cancel_evict
    test_lru.py::test_cancel_evict
    """
    ev1 = threading.Event()
    ev2 = threading.Event()

    class Slow(UserDict):
        def __getitem__(self, k):
            if when == "before":
                ev1.set()
                assert ev2.wait(timeout=5)
                return super().__getitem__(k)
            else:
                out = super().__getitem__(k)
                ev1.set()
                assert ev2.wait(timeout=5)
                return out

    buff = Buffer({}, Slow(), n=100, weight=lambda k, v: v)
    buff.set_noevict("x", 1)
    buff.fast.evict()
    assert not buff.fast
    assert set(buff.slow) == {"x"}

    with ThreadPoolExecutor(1) as ex:
        fut = ex.submit(buff.__getitem__, "x")
        assert ev1.wait(timeout=5)
        # cb is running

        if event == "set":
            buff["x"] = 2
        elif event == "set_noevict":
            buff.set_noevict("x", 2)
        else:
            assert event == "del"
            del buff["x"]
        ev2.set()

        with pytest.raises(KeyError, match="x"):
            fut.result()

    if event in ("set", "set_noevict"):
        assert buff.fast == {"x": 2}
        assert not buff.slow
        assert buff.fast.weights == {"x": 2}
        assert list(buff.fast.order) == ["x"]
    else:
        assert not buff.fast
        assert not buff.slow
        assert not buff.fast.weights
        assert not buff.fast.order

    assert not buff._cancel_restore


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
def test_stress_different_keys_threadsafe():
    # Sometimes x and y can cohexist without triggering eviction
    # Sometimes x and y are individually <n but when they're both in they cause eviction
    # Sometimes x or y are heavy
    buff = Buffer(
        {},
        utils_test.SlowDict(0.001),
        n=1,
        weight=lambda k, v: random.choice([0.4, 0.9, 1.1]),
    )
    utils_test.check_different_keys_threadsafe(buff)
    assert not buff.fast
    assert not buff.slow
    utils_test.check_mapping(buff)


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
def test_stress_same_key_threadsafe():
    # Sometimes x is heavy
    buff = Buffer(
        {},
        utils_test.SlowDict(0.001),
        n=1,
        weight=lambda k, v: random.choice([0.9, 1.1]),
    )
    utils_test.check_same_key_threadsafe(buff)
    assert not buff.fast
    assert not buff.slow
    utils_test.check_mapping(buff)



from __future__ import annotations

import threading
from collections.abc import Callable, Iterable, Iterator, Mapping
from contextlib import contextmanager
from enum import Enum
from functools import wraps
from itertools import chain
from typing import MutableMapping  # TODO move to collections.abc (needs Python >=3.9)
from typing import TYPE_CHECKING, Any, TypeVar, cast

T = TypeVar("T")
KT = TypeVar("KT")
VT = TypeVar("VT")

if TYPE_CHECKING:
    # TODO import ParamSpec from typing (needs Python >=3.10)
    # TODO import Self from typing (needs Python >=3.11)
    from typing_extensions import ParamSpec, Self

    P = ParamSpec("P")


class NoDefault(Enum):
    nodefault = None


nodefault = NoDefault.nodefault


class ZictBase(MutableMapping[KT, VT]):
    """Base class for zict mappings"""

    lock: threading.RLock

    def __init__(self) -> None:
        self.lock = threading.RLock()

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        del state["lock"]
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__ = state
        self.lock = threading.RLock()

    def update(  # type: ignore[override]
        self,
        other: Mapping[KT, VT] | Iterable[tuple[KT, VT]] = (),
        /,
        **kwargs: VT,
    ) -> None:
        if hasattr(other, "items"):
            other = other.items()
        other = chain(other, kwargs.items())  # type: ignore
        self._do_update(other)

    def _do_update(self, items: Iterable[tuple[KT, VT]]) -> None:
        # Default implementation, can be overriden for speed
        for k, v in items:
            self[k] = v

    def discard(self, key: KT) -> None:
        """Flush *key* if possible.
        Not the same as ``m.pop(key, None)``, as it doesn't trigger ``__getitem__``.
        """
        discard(self, key)

    def close(self) -> None:
        """Release any system resources held by this object"""

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()

    @contextmanager
    def unlock(self) -> Iterator[None]:
        """To be used in a method decorated by ``@locked``.
        Temporarily releases the mapping's RLock.
        """
        self.lock.release()
        try:
            yield
        finally:
            self.lock.acquire()


def close(*z: Any) -> None:
    """Close *z* if possible."""
    for zi in z:
        if hasattr(zi, "close"):
            zi.close()


def flush(*z: Any) -> None:
    """Flush *z* if possible."""
    for zi in z:
        if hasattr(zi, "flush"):
            zi.flush()


def discard(m: MutableMapping[KT, VT], key: KT) -> None:
    """Flush *key* if possible.
    Not the same as ``m.pop(key, None)``, as it doesn't trigger ``__getitem__``.
    """
    try:
        del m[key]
    except KeyError:
        pass


def locked(func: Callable[P, VT]) -> Callable[P, VT]:
    """Decorator for a method of ZictBase, which wraps the whole method in a
    instance-specific (but not key-specific) rlock.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> VT:
        self = cast(ZictBase, args[0])
        with self.lock:
            return func(*args, **kwargs)

    return wrapper



from __future__ import annotations

import asyncio
import contextvars
from collections.abc import Callable, Collection
from concurrent.futures import Executor, ThreadPoolExecutor
from functools import wraps
from itertools import chain
from typing import Any, Literal

from zict.buffer import Buffer
from zict.common import KT, VT, T, locked


class AsyncBuffer(Buffer[KT, VT]):
    """Extension of :class:`~zict.Buffer` that allows offloading all reads and writes
    from/to slow to a separate worker thread.

    This requires ``fast`` to be fully thread-safe (e.g. a plain dict).

    ``slow.__setitem__`` and ``slow.__getitem__`` will be called from the offloaded
    thread, while all of its other methods (including, notably for the purpose of
    thread-safety consideration, ``__contains__`` and ``__delitem__``) will be called
    from the main thread.

    See Also
    --------
    Buffer

    Parameters
    ----------
    Same as in Buffer, plus:

    executor: concurrent.futures.Executor, optional
        An Executor instance to use for offloading. It must not pickle/unpickle.
        Defaults to an internal ThreadPoolExecutor.
    nthreads: int, optional
        Number of offloaded threads to run in parallel. Defaults to 1.
        Mutually exclusive with executor parameter.
    """

    executor: Executor | None
    nthreads: int | None
    futures: set[asyncio.Future]
    evicting: dict[asyncio.Future, float]

    @wraps(Buffer.__init__)
    def __init__(
        self,
        *args: Any,
        executor: Executor | None = None,
        nthreads: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.executor = executor
        self.nthreads = None if executor else nthreads
        self._internal_executor = executor is None
        self.futures = set()
        self.evicting = {}

    def close(self) -> None:
        # Call LRU.close(), which stops LRU.evict_until_below_target() halfway through
        super().close()
        for future in self.futures:
            future.cancel()
        if self.executor is not None and self.nthreads is not None:
            self.executor.shutdown(wait=True)
            self.executor = None

    def _offload(self, func: Callable[..., T], *args: Any) -> asyncio.Future[T]:
        if self.executor is None:
            assert self.nthreads
            self.executor = ThreadPoolExecutor(
                self.nthreads, thread_name_prefix="zict.AsyncBuffer offloader"
            )

        loop = asyncio.get_running_loop()
        context = contextvars.copy_context()
        future = loop.run_in_executor(self.executor, context.run, func, *args)
        self.futures.add(future)
        future.add_done_callback(self.futures.remove)
        return future  # type: ignore[return-value]

    # Return an asyncio.Future, instead of just writing it as an async function, to make
    # it easier for overriders to tell apart the use case when all keys were already
    # in fast
    @locked
    def async_get(
        self, keys: Collection[KT], missing: Literal["raise", "omit"] = "raise"
    ) -> asyncio.Future[dict[KT, VT]]:
        """Fetch one or more key/value pairs. If not all keys are available in fast,
        offload to a worker thread moving keys from slow to fast, as well as possibly
        moving older keys from fast to slow.

        Parameters
        ----------
        keys:
            collection of zero or more keys to get
        missing: raise or omit, optional
            raise (default)
                If any key is missing, raise KeyError.
            omit
                If a key is missing, return a dict with less keys than those requested.

        Notes
        -----
        All keys may be present when you call ``async_get``, but ``__delitem__`` may be
        called on one of them before the actual data is fetched. ``__setitem__`` also
        internally calls ``__delitem__`` in a non-atomic way, so you may get
        ``KeyError`` when updating a value too.
        """
        # This block avoids spawning a thread if keys are missing from both fast and
        # slow. It is otherwise just a performance optimization.
        if missing == "omit":
            keys = [key for key in keys if key in self]
        elif missing == "raise":
            for key in keys:
                if key not in self:
                    raise KeyError(key)
        else:
            raise ValueError(f"missing: expected raise or omit; got {missing}")
        # End performance optimization

        try:
            # Do not pull keys towards the top of the LRU unless they are all available.
            # This matters when there is a very long queue of async_get futures.
            d = self.fast.get_all_or_nothing(keys)
        except KeyError:
            pass
        else:
            f: asyncio.Future[dict[KT, VT]] = asyncio.Future()
            f.set_result(d)
            return f

        def _async_get() -> dict[KT, VT]:
            d = {}
            for k in keys:
                if self.fast.closed:
                    raise asyncio.CancelledError()
                try:
                    # This can cause keys to be restored and older keys to be evicted
                    d[k] = self[k]
                except KeyError:
                    # Race condition: key was there when async_get was called, but got
                    # deleted afterwards.
                    if missing == "raise":
                        raise
            return d

        return self._offload(_async_get)

    def __setitem__(self, key: KT, value: VT) -> None:
        """Immediately set a key in fast. If this causes the total weight to exceed n,
        asynchronously start moving keys from fast to slow in a worker thread.
        """
        self.set_noevict(key, value)
        self.async_evict_until_below_target()

    @locked
    def async_evict_until_below_target(self, n: float | None = None) -> None:
        """If the total weight exceeds n, asynchronously start moving keys from fast to
        slow in a worker thread.
        """
        if n is None:
            n = self.n
        n = max(0.0, n)
        weight = min(chain([self.fast.total_weight], self.evicting.values()))
        if weight <= n:
            return

        # Note: this can get cancelled by LRU.close(), which in turn is
        # triggered by Buffer.close()
        future = self._offload(self.evict_until_below_target, n)
        self.evicting[future] = n
        future.add_done_callback(self.evicting.__delitem__)



import random
import threading
from collections import UserDict
from concurrent.futures import ThreadPoolExecutor

import pytest

from zict import Sieve
from zict.tests import utils_test


def test_simple():
    a = {}
    b = {}
    c = {}

    def selector(k, v):
        return len(v) % 3

    mappings = {0: a, 1: b, 2: c}

    d = Sieve(mappings, selector)
    assert len(d) == 0

    d["u"] = b"the"
    d["v"] = b"big"
    d["w"] = b"brown"
    d["x"] = b"fox"
    d["y"] = b"jumps"
    d["z"] = b"over"

    assert d["u"] == b"the"
    assert d["v"] == b"big"
    assert len(d) == 6

    assert sorted(d) == ["u", "v", "w", "x", "y", "z"]
    assert sorted(d.keys()) == ["u", "v", "w", "x", "y", "z"]
    assert sorted(d.values()) == sorted(
        [b"the", b"big", b"brown", b"fox", b"jumps", b"over"]
    )

    assert a == {"u": b"the", "v": b"big", "x": b"fox"}
    assert b == {"z": b"over"}
    assert c == {"w": b"brown", "y": b"jumps"}

    # Changing existing keys can move values from one mapping to another.
    d["w"] = b"lazy"
    d["x"] = b"dog"
    assert d["w"] == b"lazy"
    assert d["x"] == b"dog"
    assert len(d) == 6
    assert sorted(d.values()) == sorted(
        [b"the", b"big", b"lazy", b"dog", b"jumps", b"over"]
    )

    assert a == {"u": b"the", "v": b"big", "x": b"dog"}
    assert b == {"w": b"lazy", "z": b"over"}
    assert c == {"y": b"jumps"}

    del d["v"]
    del d["w"]
    assert len(d) == 4
    assert "v" not in d
    assert "w" not in d
    assert sorted(d.values()) == sorted([b"the", b"dog", b"jumps", b"over"])


def test_mapping():
    """
    Test mapping interface for Sieve().
    """
    a = {}
    b = {}

    def selector(key, value):
        return sum(bytearray(value)) & 1

    mappings = {0: a, 1: b}
    z = Sieve(mappings, selector)
    utils_test.check_mapping(z)
    utils_test.check_closing(z)

    z.clear()
    assert z.mappings == {0: {}, 1: {}}
    assert not z.key_to_mapping


@pytest.mark.parametrize("method", ("__setitem__", "update"))
@pytest.mark.parametrize("set_when", ("before", "after"))
@pytest.mark.parametrize("seed", [False, "same", "different"])
def test_multithread_race_condition_del_set(method, set_when, seed):
    """Test race conditions between __delitem__ and __setitem__/update on the same key"""
    in_set = threading.Event()
    block_set = threading.Event()

    class Slow(UserDict):
        def __setitem__(self, k, v):
            if set_when == "before":
                self.data[k] = v
                in_set.set()
                assert block_set.wait(timeout=5)
            else:
                in_set.set()
                assert block_set.wait(timeout=5)
                self.data[k] = v

    z = Sieve({0: {}, 1: Slow()}, selector=lambda k, v: v % 2)
    if seed == "same":
        block_set.set()
        z["x"] = 1  # mapping 1
        in_set.clear()
        block_set.clear()
    elif seed == "different":
        z["x"] = 0  # mapping 0

    with ThreadPoolExecutor(1) as ex:
        if method == "__setitem__":
            set_fut = ex.submit(z.__setitem__, "x", 3)  # mapping 1
        else:
            assert method == "update"
            set_fut = ex.submit(z.update, {"x": 3})  # mapping 1
        assert in_set.wait(timeout=5)
        try:
            del z["x"]
        except KeyError:
            pass
        block_set.set()
        set_fut.result()

    assert not z.mappings[0]
    assert not z.mappings[1]
    assert not z.key_to_mapping


@pytest.mark.parametrize("set1_when", ("before", "after"))
@pytest.mark.parametrize("set2_when", ("before", "after"))
@pytest.mark.parametrize("set1_method", ("__setitem__", "update"))
@pytest.mark.parametrize("set2_method", ("__setitem__", "update"))
@pytest.mark.parametrize("starts_first", (1, 2))
@pytest.mark.parametrize("ends_first", (1, 2))
@pytest.mark.parametrize("seed", [False, 0, 1, 2])
def test_multithread_race_condition_set_set(
    set1_when, set2_when, set1_method, set2_method, starts_first, ends_first, seed
):
    """Test __setitem__/update in race condition with __setitem__/update"""
    when = {1: set1_when, 2: set2_when}
    method = {1: set1_method, 2: set2_method}
    in_set = {1: threading.Event(), 2: threading.Event()}
    block_set = {1: threading.Event(), 2: threading.Event()}

    class Slow(UserDict):
        def __setitem__(self, k, v):
            if v < 3:
                # seed
                self.data[k] = v
                return
            mkey = v % 3

            if when[mkey] == "before":
                self.data[k] = v
                in_set[mkey].set()
                assert block_set[mkey].wait(timeout=5)
            else:
                in_set[mkey].set()
                assert block_set[mkey].wait(timeout=5)
                self.data[k] = v

    z = Sieve({0: {}, 1: Slow(), 2: Slow()}, selector=lambda k, v: v % 3)
    if seed is not False:
        z["x"] = seed

    with ThreadPoolExecutor(2) as ex:
        futures = {}
        starts_second = 2 if starts_first == 1 else 1
        for idx in (starts_first, starts_second):
            if method[idx] == "__setitem__":
                futures[idx] = ex.submit(z.__setitem__, "x", idx + 3)
            else:
                assert method[idx] == "update"
                futures[idx] = ex.submit(z.update, {"x": idx + 3})
            assert in_set[idx].wait(timeout=5)

        block_set[ends_first].set()
        futures[ends_first].result()
        ends_second = 2 if ends_first == 1 else 1
        block_set[ends_second].set()
        futures[ends_second].result()

    assert dict(z) in ({"x": 4}, {"x": 5})
    assert z.mappings[0] == {}
    if z["x"] == 4:
        assert z.mappings[1] == {"x": 4}
        assert z.mappings[2] == {}
        assert z.key_to_mapping == {"x": z.mappings[1]}
    else:
        assert z.mappings[1] == {}
        assert z.mappings[2] == {"x": 5}
        assert z.key_to_mapping == {"x": z.mappings[2]}


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
def test_stress_different_keys_threadsafe():
    a = {}
    b = {}
    z = Sieve({0: a, 1: b}, lambda k, v: random.choice([0, 1]))
    utils_test.check_different_keys_threadsafe(z)
    assert not a
    assert not b
    assert not z.key_to_mapping
    utils_test.check_mapping(z)


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
def test_stress_same_key_threadsafe():
    a = utils_test.SlowDict(0.001)
    b = utils_test.SlowDict(0.001)
    z = Sieve({0: a, 1: b}, lambda k, v: random.choice([0, 1]))
    utils_test.check_same_key_threadsafe(z)
    assert not a
    assert not b
    assert not z.key_to_mapping
    utils_test.check_mapping(z)



from __future__ import annotations

from collections.abc import Iterable, Iterator
from typing import MutableSet  # TODO import from collections.abc (needs Python >=3.9)

from zict.common import T


class InsertionSortedSet(MutableSet[T]):
    """A set-like that retains insertion order, like a dict. Thread-safe.

    Equality does not compare order or class, but only compares against the contents of
    any other set-like, coherently with dict and the AbstractSet design.
    """

    _d: dict[T, None]
    __slots__ = ("_d",)

    def __init__(self, other: Iterable[T] = ()) -> None:
        self._d = dict.fromkeys(other)

    def __contains__(self, item: object) -> bool:
        return item in self._d

    def __iter__(self) -> Iterator[T]:
        return iter(self._d)

    def __len__(self) -> int:
        return len(self._d)

    def add(self, value: T) -> None:
        """Add element to the set. If the element is already in the set, retain original
        insertion order.
        """
        self._d[value] = None

    def discard(self, value: T) -> None:
        # Don't trust the thread-safety of self._d.pop(value, None)
        try:
            del self._d[value]
        except KeyError:
            pass

    def remove(self, value: T) -> None:
        del self._d[value]

    def popleft(self) -> T:
        """Pop the oldest-inserted key from the set"""
        while True:
            try:
                value = next(iter(self._d))
                del self._d[value]
                return value
            except StopIteration:
                raise KeyError("pop from an empty set")
            except (KeyError, RuntimeError):
                # Multithreaded race condition
                continue

    def popright(self) -> T:
        """Pop the latest-inserted key from the set"""
        return self._d.popitem()[0]

    pop = popright

    def clear(self) -> None:
        self._d.clear()
